#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

//NOT a 4d array with page, row, col, etc.
//but a 2d array with 4 cols
#define ASTRUCT_MAXROWS 1
#define ASTRUCT_MAXCOLS 1 //was 3

extern double Mygsl_rhy_dists(int distribnumber, int counter, const gsl_rng * r);


struct the4dstruct
{
    int structnum;
    int nrows;
    int ncols;
    float a4dstruct[ASTRUCT_MAXROWS][ASTRUCT_MAXCOLS];
};


//------------- I N F O ------------------
//function takes in a two dimensional array, with four cols
//and the number of tetrads to be returned
//
//2d array tetrads are each loaded into a struct
//ex. (0,0,0,0), (0,0,0,1). (0,0,0,2), and (0,0,0,3)
//
//then an array of those structs is created
//then randomized, returning a 2 dim array, with 4 cols
//of the number of tetrads
//------------- E N D  I N F O ------------------

//NOT a 4d array with page, row, col, etc.
//but a 2d array with 4 cols
float *random4dstructs(int inrows, int incols, float **input4darray,
    int returnrows, int returncols, int rhydistrib)
{
    char c;
    //mem ptrs for ran gen
    const gsl_rng_type * T;
    gsl_rng * r;

    int spinrows = returnrows;
    int spincols = returncols;
    int spincount = returnrows * returncols;
    int hh = inrows * incols;
    int h; //highest number
    int l = 1; //lowest number
    int f, i, j, k, nn, x, y, forthnumrange;
    int rhydist = rhydistrib;
    int distcounter;

    float **s = NULL;  //the local 2d array
    float *sPool, *curS_Ptr; //for calloc
    float *ss = NULL; //local 1d arrays
    int *pp = NULL;
    int *tt = NULL;

    double u; //value returned from Mygsl_rhy_dists

    struct the4dstruct **Arrayof4dStructs = NULL;
    struct timeval  t1; //structs for rand using time

first:
    distcounter = 1;  //sets or resets counter
//*****************************************
//********  ALLOCATE MEMORY ***************
//*****************************************
    /* create a generator chosen by the
       environment variable GSL_RNG_TYPE.
       In this case using 'taus' algo */
    gsl_rng_env_setup();
    T = gsl_rng_taus; //use taus algo as ran gen
    r = gsl_rng_alloc (T);  //alloc mem, see free below

//allocate local initialized 2d array memory, with 4 cols.
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    s=(float **) calloc(spinrows+1, sizeof(float*));
    sPool = (float*)calloc(spincols*spinrows,sizeof(float));
    curS_Ptr=sPool;
    for (i = 0; i<spinrows+1; i++)
    {
        s[i]= curS_Ptr;
        curS_Ptr += spincols;
    }

    //allocate local initialized 1d array memory
    ss=(float *) calloc(spincount+1, sizeof(float*));
    pp=(int *) calloc(spincount, sizeof(int*));
    tt=(int *) calloc(spincount, sizeof(int*));


//allocate memory for one 4d struct
    Arrayof4dStructs = (struct the4dstruct *)calloc(1, sizeof(struct the4dstruct));
//allocate memory for the inrows*incols num of 4dstructs
    h= inrows*incols;
    for (x=0; x<=h; x++)
    {
        /* add ONE element to the array */
        Arrayof4dStructs= (struct the4dstruct **)realloc(Arrayof4dStructs, (x + 1) * sizeof(struct the4dstruct *));
        /* allocate memory for x number of the4dstructs */
        Arrayof4dStructs[x] = (struct the4dstruct *)malloc(sizeof(struct the4dstruct));
    }

//load data from incoming 2d array to local array of structs
//structnums begin with '1' for ease of counting and printing,
//but structs begin with '0'
    x=0;
    y=1;
    for (j = 0; j<inrows; j++)
    {
        for (k = 0; k<incols; k++)
        {
            Arrayof4dStructs[x]->a4dstruct[0][0] = input4darray[j][k];
            Arrayof4dStructs[x]->structnum = y;
            x++;
            y++;
        }
    }

//load every fourth number into local 1d array (1,5,9,13,17,21...)
//hh=inrows, incols
    forthnumrange=1;
    for (x=1; x<=hh; x+=4)
    {
        pp[forthnumrange] = x;
        forthnumrange++;
    }


//*****************************************************************
//********  RANDOMIZE array of structures (arcsine) ***************
//*****************************************************************

    h = forthnumrange;

    for (x=1; x<=spincount/4; x++)
    {
next:
        gettimeofday(&t1, NULL);
        gsl_rng_default_seed = t1.tv_usec * t1.tv_sec; //use time as gen seed

        //****GSL. Get a random distribution****
        u = Mygsl_rhy_dists(rhydist, distcounter, r);

        f = (int)(u*h-1)+1; //limits between ranges

        if (distcounter > 3000)
        {
            free (Arrayof4dStructs);
            free(*s);
            free(s);
            free(ss);
            free(pp);
            free(tt);
            gsl_rng_free (r);
            printf("\n The distribution is not generating numbers\
                \n within the ranges specified.  Try entering\
                \n different input values for the distribution. \n");
            goto first;
        }

        if(h == 1)   //avoids rare case where highest = 1
        {
            if(f<l || f>h)
            {
                distcounter++;
                goto next;
            }
        }
        else
        {
            if(f<l || f>h-1)  /*avoid out of bounds nums */
            {
                distcounter++;
                goto next;
            }
        }

        tt[x]= pp[f];
        distcounter++;
    } //end for loop
//+++++++++++++++++++++

//*********************************************************
//********  LOADING and PRINTING spun array ***************
//*********************************************************
    x=0;
    for(i=1; i<=spincount/4; i++)
    {
        nn=tt[i];
        x=x+1;
        ss[x]=Arrayof4dStructs[nn-1]->a4dstruct[0][0]; //1, 5, 9, 13, 17, 21 ...
        x=x+1;
        ss[x]=Arrayof4dStructs[nn]->a4dstruct[0][0];   //2, 6, 10, 14, 18, 22...
        x=x+1;
        ss[x]=Arrayof4dStructs[nn+1]->a4dstruct[0][0]; //3, 7, 11, 15, 19, 23...
        x=x+1;
        ss[x]=Arrayof4dStructs[nn+2]->a4dstruct[0][0]; //4, 8, 12, 16, 20, 24...
    }

// loads the 1d array into a 2d array to return to the
// RHY4DINPUT func below
    printf("\n...printing the spincount of 2d spun structs (from rand4dstructs)...\n");
    i=0;
    for (j = 0; j<spinrows; j++)
    {
        for (k = 0; k<spincols; k++)
        {
            i++;
            s[j][k] = ss[i];
             printf(" %4.6f ", s[j][k]);
        }
    }

//should be a switch here to do another spin before returning values
    printf("\ntry again? (y)es, (n)o: ");
    while (getchar()!='\n');
    scanf("%c", &c);
    switch(c)
    {
    case 'y':
        //free ALL memory allocated so far
        for (x=0; x<=h; x++)
        {
            free(Arrayof4dStructs[x]);
        }
        free (Arrayof4dStructs);
        free(*s);//NOTE: this how ptr pool helps with freeing memory
        free(s);
        free(ss);
        free(pp);
        free(tt);
        gsl_rng_free (r);
        goto first;
        break;
    case 'n':
        printf("\nbye");
        break;
    default:
        break;
    }//ends switch
    printf("\n");

    for (x=0; x<=h; x++)
    {
        free(Arrayof4dStructs[x]);
    }
    free (Arrayof4dStructs);
    free(ss);
    free(pp);
    free(tt);
    gsl_rng_free (r);

    return s;

} //end random4dstructs

//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++

//******   RHY4DINPUT    ******

//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++

//NOT a 4d array with page, row, col
//but a 2d array with 4 cols
float *rhy4Dinput_gsl(float *rhy4Dreturnarray)
{
    char c;
    float **struct4dinputarray = NULL; //local 2d array with 4 cols
    float **struct4dreturnarray = NULL; //locadl 2d array with 4 cols
    float  *struct1dreturnarray = NULL;
    int i, j, k, nn, x;
    int rows, spinrows;  //number of input rows and returned rows
    int cols = 4;  //one input pair (for 4d leave cols as "4")
    int spincols = 4; //one returned group of three
    float *inputPool, *returnPool; //for calloc
    float *curInputPtr, *curReturnPtr; //for calloc

    int rhydist;//the distribution selection

//**************************************
//********  ENTER VALUES ***************
//**************************************
//fill a 2d array, with 4 cols, to send to rand3dsturcts function
start:
    printf("\n");
    printf("Enter the number of tetrads: ");
    scanf("%d",&rows);

//allocate regular 2d array memory
//keep 2d array as 3 cols
//But rows * num of pairs are needed for input
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    struct4dinputarray=(float **) calloc(rows, sizeof(float*));
    inputPool = (float*)calloc(cols*rows,sizeof(float));
    curInputPtr=inputPool;
    for (i = 0; i<rows; i++)
    {
        struct4dinputarray[i]= curInputPtr;
        curInputPtr += cols;
    }

//load the 2d array with rhythmic tetrads, ex:
//(0,0)=0.25, (0,1)=0.25, (0,2)=0.25, (0,3)=0.25
//(1,0)=0.5, (1,1)=0.5, (1,2)=0.5, (1,3)=0.5, etc.
    x=1;
    for (j = 0; j<rows; j++)
    {
        for (k = 0; k<cols; k++)
        {
            printf("Tetrad #%d, enter value #%d: ", j+1, x);
            scanf("%f",&struct4dinputarray [j][k]);
            x++;
        }
    }
    //------------- I N F O ------------------
// todo:  add error checking here to make sure each group of four
// sums to a whole number.  In that way "Beats" returned
// will be clear.
// However could also just leave it up to the user to manage the
// the entry of values, so for example syncopation, or groups
// of three larger than one beat(quarter) could result
//------------- E N D  I N F O ------------------

    printf("\n...here's the list \n");
    for (j = 0; j<rows; j++)
    {
        for (k = 0; k<cols; k++)
        {
            printf("%2.4f\n",  struct4dinputarray[j][k]); // Displaying the Array
        }
    }

//switch for another spin before returning values
    printf("\ntry again? (y)es, (n)o: ");
    while (getchar()!='\n');
    scanf("%c", &c);
    switch(c)
    {
    case 'y':
        //free 2d array memory
//NOTE: this how ptr pool helps with freeing memory
        free(*struct4dinputarray);
        free(struct4dinputarray);
        goto start;
        break;
    case 'n':
        goto next;
        break;
    }//ends switch
    printf("\n");
//**************************************
//********  RETURN VALUES ***************
//**************************************
next:
//number of groups of four to be returned
    printf("\nEnter the number of tetrads to be returned: ");
    scanf("%d",&spinrows);

//allocate 1d array memory
    struct1dreturnarray=(float **) calloc((spinrows*spincols), sizeof(float*));

//allocate 2d array memory, with four cols
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    struct4dreturnarray=(float **) calloc(spinrows, sizeof(float*));
    returnPool = (float*)calloc(spincols*spinrows,sizeof(float));
    curReturnPtr = returnPool;
    for (i = 0; i<spinrows; i++)
    {
        struct4dreturnarray[i]= curReturnPtr;
        curReturnPtr += spincols;
    }


//***************************************
//*****  SELECT RANDOM DISTRIBUTION *****
//***************************************

    printf("\n");
    printf("\t\t\t***** RHYTHM DISTRIBUTIONS *****");
    printf("\n1 = gaussian, 2 = gaussian_tail, 3 = exponential, 4 = laplace\
            \n5 = exppow, 6 = cauchy, 7 = rayleigh, 8 = rayleigh_tail\
            \n9 = landau, 10 = levy, 11 = levy_skew, 12 = gamma\
            \n13 = flat, 14 = lognormal, 15 = chisq, 16 = fdist\
            \n17 = tdist, 18 = beta, 19 = logistic, 20 = pareto\
            \n21 = Weibell, 22 = gumbel1, 23 = gumbel2, 24 = poisson\
            \n25 = binomial, 26 = negative_binomial, 27 = pascal\
            \n28 = geometric, 29 = hypergeometric, 30 = logarithmic\
            \n\nEnter a distribution for rhythm:  ");
    scanf("%d",&rhydist);

//***************************************
//********  CALL random2dstructs_gsl ********
//***************************************

    struct4dreturnarray = random4dstructs(rows, cols, struct4dinputarray, spinrows, spincols, rhydist);

    x=0;
    for (j = 0; j<spinrows; j++)
    {
        for (k = 0; k<spincols; k++)
        {
            x++;
            struct1dreturnarray[x] = struct4dreturnarray[j][k];
        }
    }

    //free 2d array memory
//NOTE: this how ptr pool helps with freeing memory
    free(*struct4dinputarray);
    free(struct4dinputarray);

    free(*struct4dreturnarray);
    free(struct4dreturnarray);

    return struct1dreturnarray;   //returns a 1d array of rhythms

} //end rhy4dinput
